@@ -1,6 +1,8 @@
 import asyncio
+import asyncio.coroutines
 import functools
 import os
+import sys
 import threading
 from concurrent.futures import Future, ThreadPoolExecutor
 
@@ -13,6 +15,17 @@ except ImportError:
     contextvars = None
 
 
+def _restore_context(context):
+    # Check for changes in contextvars, and set them to the current
+    # context for downstream consumers
+    for cvar in context:
+        try:
+            if cvar.get() != context.get(cvar):
+                cvar.set(context.get(cvar))
+        except LookupError:
+            cvar.set(context.get(cvar))
+
+
 class AsyncToSync:
     """
     Utility class which turns an awaitable that only works on the thread with
@@ -35,6 +48,10 @@ class AsyncToSync:
 
     def __init__(self, awaitable, force_new_loop=False):
         self.awaitable = awaitable
+        try:
+            self.__self__ = self.awaitable.__self__
+        except AttributeError:
+            pass
         if force_new_loop:
             # They have asked that we always run in a new sub-loop.
             self.main_event_loop = None
@@ -60,6 +77,14 @@ class AsyncToSync:
                     "You cannot use AsyncToSync in the same thread as an async event loop - "
                     "just await the async function directly."
                 )
+
+        if contextvars is not None:
+            # Wrapping context in list so it can be reassigned from within
+            # `main_wrap`.
+            context = [contextvars.copy_context()]
+        else:
+            context = None
+
         # Make a future for the return information
         call_result = Future()
         # Get the source thread
@@ -77,14 +102,16 @@ class AsyncToSync:
         # main event loop's thread if it's there, otherwise make a new loop
         # in this thread.
         try:
+            awaitable = self.main_wrap(
+                args, kwargs, call_result, source_thread, sys.exc_info(), context
+            )
+
             if not (self.main_event_loop and self.main_event_loop.is_running()):
                 # Make our own event loop - in a new thread - and run inside that.
                 loop = asyncio.new_event_loop()
                 loop_executor = ThreadPoolExecutor(max_workers=1)
                 loop_future = loop_executor.submit(
-                    self._run_event_loop,
-                    loop,
-                    self.main_wrap(args, kwargs, call_result, source_thread),
+                    self._run_event_loop, loop, awaitable
                 )
                 if current_executor:
                     # Run the CurrentThreadExecutor until the future is done
@@ -94,8 +121,7 @@ class AsyncToSync:
             else:
                 # Call it inside the existing loop
                 self.main_event_loop.call_soon_threadsafe(
-                    self.main_event_loop.create_task,
-                    self.main_wrap(args, kwargs, call_result, source_thread),
+                    self.main_event_loop.create_task, awaitable
                 )
                 if current_executor:
                     # Run the CurrentThreadExecutor until the future is done
@@ -106,6 +132,9 @@ class AsyncToSync:
                 del self.executors.current
             if old_current_executor:
                 self.executors.current = old_current_executor
+            if contextvars is not None:
+                _restore_context(context[0])
+
         # Wait for results from the future.
         return call_result.result()
 
@@ -118,6 +147,26 @@ class AsyncToSync:
             loop.run_until_complete(coro)
         finally:
             try:
+                # mimic asyncio.run() behavior
+                # cancel unexhausted async generators
+                if sys.version_info >= (3, 7, 0):
+                    tasks = asyncio.all_tasks(loop)
+                else:
+                    tasks = asyncio.Task.all_tasks(loop)
+                for task in tasks:
+                    task.cancel()
+                loop.run_until_complete(asyncio.gather(*tasks, return_exceptions=True))
+                for task in tasks:
+                    if task.cancelled():
+                        continue
+                    if task.exception() is not None:
+                        loop.call_exception_handler(
+                            {
+                                "message": "unhandled exception during loop shutdown",
+                                "exception": task.exception(),
+                                "task": task,
+                            }
+                        )
                 if hasattr(loop, "shutdown_asyncgens"):
                     loop.run_until_complete(loop.shutdown_asyncgens())
             finally:
@@ -131,15 +180,28 @@ class AsyncToSync:
         func = functools.partial(self.__call__, parent)
         return functools.update_wrapper(func, self.awaitable)
 
-    async def main_wrap(self, args, kwargs, call_result, source_thread):
+    async def main_wrap(
+        self, args, kwargs, call_result, source_thread, exc_info, context
+    ):
         """
         Wraps the awaitable with something that puts the result into the
         result/exception future.
         """
+        if context is not None:
+            _restore_context(context[0])
+
         current_task = SyncToAsync.get_current_task()
         self.launch_map[current_task] = source_thread
         try:
-            result = await self.awaitable(*args, **kwargs)
+            # If we have an exception, run the function inside the except block
+            # after raising it so exc_info is correctly populated.
+            if exc_info[1]:
+                try:
+                    raise exc_info[1]
+                except:
+                    result = await self.awaitable(*args, **kwargs)
+            else:
+                result = await self.awaitable(*args, **kwargs)
         except Exception as e:
             call_result.set_exception(e)
         else:
@@ -147,6 +209,9 @@ class AsyncToSync:
         finally:
             del self.launch_map[current_task]
 
+            if context is not None:
+                context[0] = contextvars.copy_context()
+
 
 class SyncToAsync:
     """
@@ -184,7 +249,13 @@ class SyncToAsync:
 
     def __init__(self, func, thread_sensitive=False):
         self.func = func
+        functools.update_wrapper(self, func)
         self._thread_sensitive = thread_sensitive
+        self._is_coroutine = asyncio.coroutines._is_coroutine
+        try:
+            self.__self__ = func.__self__
+        except AttributeError:
+            pass
 
     async def __call__(self, *args, **kwargs):
         loop = asyncio.get_event_loop()
@@ -216,12 +287,18 @@ class SyncToAsync:
                 self.thread_handler,
                 loop,
                 self.get_current_task(),
+                sys.exc_info(),
                 func,
                 *args,
                 **kwargs
             ),
         )
-        return await asyncio.wait_for(future, timeout=None)
+        ret = await asyncio.wait_for(future, timeout=None)
+
+        if contextvars is not None:
+            _restore_context(context)
+
+        return ret
 
     def __get__(self, parent, objtype):
         """
@@ -229,7 +306,7 @@ class SyncToAsync:
         """
         return functools.partial(self.__call__, parent)
 
-    def thread_handler(self, loop, source_task, func, *args, **kwargs):
+    def thread_handler(self, loop, source_task, exc_info, func, *args, **kwargs):
         """
         Wraps the sync application with exception handling.
         """
@@ -246,7 +323,15 @@ class SyncToAsync:
             parent_set = True
         # Run the function
         try:
-            return func(*args, **kwargs)
+            # If we have an exception, run the function inside the except block
+            # after raising it so exc_info is correctly populated.
+            if exc_info[1]:
+                try:
+                    raise exc_info[1]
+                except:
+                    return func(*args, **kwargs)
+            else:
+                return func(*args, **kwargs)
         finally:
             # Only delete the launch_map parent if we set it, otherwise it is
             # from someone else.
